# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Main configuration package."""

__all__ = [
    '__version__',
    'Configuration', 'Category', 'Parameter', 'configuration', 'category',
    'BOOL', 'ARRAY', 'PType',
    'Configurable', 'applyconfiguration',
    'ConfDriver'
]

from .version import __version__
from .configurable import Configurable, applyconfiguration
from .model import (
    Configuration, Category, Parameter, configuration, category, BOOL, ARRAY,
    PType
)
from .driver import ConfDriver



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Resolver registry.

This module aims to manage several resolvers in a registry.

This module defines expression resolvers for dedicated programming languages.

An expression resolver is a function which takes in parameters:

- expr: string expression to resolve.
- safe: boolean flag for a safe execution context.
- tostr: boolean flag about format the expression into a string representation.
- scope: dict of variables for execution context (contain variables).

A resolver is registered with the function ``register``.
And is loaded in three ways:

- in setting the environment variable 'B3J0F_EXPRRES_PATH' (for example,
    `b3j0f.conf.parser.lang.js,custom.c` could load both modules `js` and `c`
    containing programming language parsers).
- in using the function `loadresolvers`.
- in simply importing a dedicated module with the import keyword.
"""

__all__ = [
    'ResolverRegistry',
    'names', 'resolve', 'register', 'unregister', 'loadresolvers',
    'defaultname', 'getname'
]

from b3j0f.utils.path import lookup
from b3j0f.utils.version import OrderedDict

from inspect import isclass

from six import string_types

from os import getenv

from .core import (
    DEFAULT_BESTEFFORT, DEFAULT_SAFE, DEFAULT_TOSTR, DEFAULT_SCOPE
)


def loadresolvers(paths):
    """Load input path modules in order to register external resolvers.

    :param list paths: list of paths to load."""

    for path in paths:
        lookup(path)

#: env variable name of exprres modules.
B3J0F_EXPRRES_PATH = 'B3J0F_EXPRRES_PATH'

EXPRRES_PATH = getenv(B3J0F_EXPRRES_PATH)  #: expression resolver modules.

if EXPRRES_PATH:  # load modules
    loadresolvers(EXPRRES_PATH.split(','))

__RESOLVER__ = '__resolver__'  #: resolver name for class IoC


class ResolverRegistry(OrderedDict):
    """Resolver registry."""

    __slots__ = ('_default')

    def __init__(self, default=None, **resolvers):

        super(ResolverRegistry, self).__init__(resolvers)

        self._default = None

        self.default = default

    @property
    def default(self):
        """Get default resolver name."""

        if self._default is None or self._default not in self:
            self._default = list(self.keys())[0] if self else None

        return self._default

    @default.setter
    def default(self, value):
        """Change of resolver name.

        :param value: new default value to use.
        :type value: str or callable
        :raises: KeyError if value is a string not already registered."""

        if value is None:
            if self:
                value = list(self.keys())[0]

        elif not isinstance(value, string_types):
            value = register(exprresolver=value, reg=self)

        elif value not in self:
            raise KeyError(
                '{0} not registered in {1}'.format(value, self)
            )

        self._default = value

    def resolve(
            self, expr, name,
            safe=DEFAULT_SAFE, tostr=DEFAULT_TOSTR, scope=DEFAULT_SCOPE,
            besteffort=DEFAULT_BESTEFFORT
    ):
        """Resolve an expression with possibly a dedicated expression resolvers.

        :param str name: expression resolver registered name. Default is the
            first registered expression resolver.
        :param bool safe: if True (Default), resolve in a safe context
            (without I/O).
        :param bool tostr: if True (False by default), transform the result into
            a string format.
        :param dict scope: scope execution resolution.
        :return: a string if tostr, otherwise, a python object.

        :raises: KeyError if no expression resolver has been registered or if
            name does not exist in expression resolvers.
        """

        resolver = None

        if name is None:
            name = self.default

        resolver = self[name]

        result = resolver(
            expr=expr,
            safe=safe, tostr=tostr, scope=scope, besteffort=besteffort
        )

        return result

_RESOLVER_REGISTRY = ResolverRegistry()  #: default registry


def register(name=None, exprresolver=None, params=None, reg=None):
    """Register an expression resolver.

    Can be used such as a decorator.

    For example all remainding expressions are the same.

    .. code-block:: python

        @register('myresolver')
        def myresolver(**kwargs): pass

    .. code-block:: python

        def myresolver(**kwargs): pass

        register('myresolver', myresolver)

    .. code-block:: python

        @register('myresolver')
        class MyResolver():
            def __call__(**kwargs): pass

    .. code-block:: python

        class MyResolver():
            def __call__(**kwargs): pass

        register('myresolver', MyResolver)

    .. code-block:: python

        class MyResolver():
            def __call__(**kwargs): pass

        register('myresolver', MyResolver())


    :param str name: register name to use. Default is the function/class
        name.
    :param exprresolver: function able to resolve an expression.
    :param dict kwargs: parameters of resolver instanciation if exprresolver
        is a class and this function is used such as a decorator.
    :param ResolverRegistry reg: registry to use. Default is the global registry
    .
    """

    def _register(exprresolver, _name=name, _params=params):
        """Local registration for better use in a decoration context.

        :param exprresolver: function/class able to resolve an expression.
        :param str _name: private parameter used to set the name.
        :param dict _params: private parameter used to set resolver class
            constructor parameters."""

        _exprresolver = exprresolver

        if isclass(exprresolver):  # try to instanciate exprresolver
            if _params is None:
                _params = {}

            _exprresolver = _exprresolver(**_params)

        if _name is None:
            _name = getname(exprresolver)

        reg[_name] = _exprresolver

        if reg.default is None:
            reg.default = _name

        return exprresolver

    if name is None:
        name = getname(exprresolver)

    if reg is None:
        reg = _RESOLVER_REGISTRY

    if exprresolver is None:
        result = _register

    else:
        result = name
        _register(exprresolver)

    return result


def unregister(name):
    """Unregister resolver.

    :param str name: resolver name to unregister.
    :return: named resolver.
    :rtype: callable"""

    return _RESOLVER_REGISTRY.pop(name)


def defaultname(name=None):
    """Get default resolver name.

    :param str name: if given, change the default name.
    :raises: NameError if name given and not in registered resolvers.
    :rtype: str"""

    if name is None:
        name = _RESOLVER_REGISTRY.default

    else:
        _RESOLVER_REGISTRY.default = name

    return name

def names():
    """Get resolver names.

    :rtype: list"""

    return list(_RESOLVER_REGISTRY.keys())

def resolve(
        expr, name=None,
        safe=DEFAULT_SAFE, tostr=DEFAULT_TOSTR, scope=DEFAULT_SCOPE,
        besteffort=DEFAULT_BESTEFFORT
):
    """Resolve an expression with possibly a dedicated expression resolvers.

    :param str name: expression resolver registered name. Default is the
        first registered expression resolver.
    :param bool safe: if True (Default), resolve in a safe context
        (without I/O).
    :param bool tostr: if True (False by default), transform the result into
        a string format.
    :param dict scope: scope execution resolution.
    :param bool besteffort: if True (default) auto import unknown var.

    :raises: KeyError if no expression resolver has been registered or if
        name does not exist in expression resolvers.
    """

    return _RESOLVER_REGISTRY.resolve(
        expr=expr, name=name, safe=safe, tostr=tostr, scope=scope,
        besteffort=besteffort
    )

def getname(exprresolver):
    """Get expression resolver name.

    Expression resolver name is given by the attribute __resolver__.
    If not exist, then the it is given by the attribute __name__.
    Otherwise, given by the __class__.__name__ attribute.

    :raises: TypeError if exprresolver is not callable."""

    result = None

    if not callable(exprresolver):
        raise TypeError('Expression resolver must be a callable object.')

    result = getattr(
        exprresolver, __RESOLVER__, getattr(
            exprresolver, '__name__', getattr(
                exprresolver.__class__, '__name__'
            )
        )
    )

    return result



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""core resolver module.

Contains signature of the resolver function and and default parameter values."""

__all__ = [
    'DEFAULT_BESTEFFORT', 'DEFAULT_SAFE', 'DEFAULT_TOSTR', 'DEFAULT_SCOPE',
    'resolver'
]

DEFAULT_BESTEFFORT = True  #: default best effort execution context flag.
DEFAULT_SAFE = True  #: default safe execuction context flag.
DEFAULT_TOSTR = False  #: default conversion to str flag.
DEFAULT_SCOPE = None  #: default scope execution context.

def resolver(
        expr, safe=DEFAULT_SAFE, tostr=DEFAULT_TOSTR, scope=DEFAULT_SCOPE,
        besteffort=DEFAULT_BESTEFFORT
):
    """Resolve input expression.

    This function is given such as template resolution function. For wrapping
    test for example.

    :param str expr: configuration expression to resolve.
    :param bool safe: if True (default), run safely execution context.
    :param bool tostr: format the result.
    :param dict scope: execution scope (contains references to expression
        objects).
    :param bool besteffort: if True (default), try to resolve unknown variable
        name with execution runtime.
    :return: resolved expression."""


    raise NotImplementedError()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Ini configuration file driver."""

__all__ = ['INIFileConfDriver']

from six.moves.configparser import RawConfigParser

from .base import FileConfDriver
from ...model.param import Parameter


class INIFileConfDriver(FileConfDriver):
    """Manage ini resource configuration."""

    def resource(self):

        return RawConfigParser()

    def _pathresource(self, rscpath):

        result = RawConfigParser()

        result.read(rscpath)

        if not result.sections():
            result = None

        return result

    def _cnames(self, resource):

        return resource.sections()

    def _params(self, resource, cname):

        return list(
            Parameter(item[0], svalue=item[1]) for item in resource.items(cname)
        )

    def _setconf(self, conf, resource, rscpath):

        for category in conf.values():

            if not resource.has_section(category.name):
                resource.add_section(category.name)

            for param in category.values():
                resource.set(category.name, param.name, param.svalue)

        with open(rscpath, 'w') as fps:
            resource.write(fps)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""model.cat UTs."""

from unittest import main

from b3j0f.utils.ut import UTCase

from ..cat import Category
from ..param import Parameter


class CategoryTest(UTCase):
    """Test exprparser."""

    def setUp(self):

        self.name = 'test'
        self.cat = Category(name=self.name)

        for i in range(1, len(self.name)):
            param = Parameter(self.name[:i])
            rparam = Parameter('^{0}.*'.format(self.name[:i]))
            self.cat += param, rparam

    def test_name(self):
        """Test name."""

        self.assertEqual(self.cat.name, self.name)

    def test_getparams(self):
        """Test the method getparams."""

        param = Parameter(self.name)

        params = self.cat.getparams(param=param)

        self.assertEqual(len(self.name) - 1, len(params))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""JSON configuration driver."""

from __future__ import absolute_import

__all__ = ['JSONConfDriver']

try:
    from json import loads, dumps

except ImportError:
    from simplejson import loads, dumps

from .base import ConfDriver
from ..model.param import Parameter


class JSONConfDriver(ConfDriver):
    """Manage json resource configuration."""

    def rscpaths(self, path):

        return [path]

    def resource(self):

        return {}

    def _pathresource(self, rscpath):

        result = loads(rscpath)

        return result

    def _cnames(self, resource):

        return resource.keys()

    def _params(self, resource, cname):

        params = resource[cname]

        result = [
            Parameter(name=key, svalue=params[key]) for key in params
        ]

        return result

    def _setconf(self, conf, resource, rscpath):

        for category in conf.values():

            cat = resource.setdefault(category.name, {})

            for parameter in category.values():

                cat[parameter.name] = parameter.svalue

        result = dumps(resource)

        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""conf file driver UTs."""

from b3j0f.utils.ut import UTCase

from unittest import main

from ...model.conf import Configuration
from ...model.cat import Category
from ...model.param import Parameter

from ..base import ConfDriver


class TestConfDriver(ConfDriver):
    """Configuration driver for test."""

    def __init__(self):

        super(TestConfDriver, self).__init__()

        self.confbypath = {}

    def resource(self):

        return {}

    def _pathresource(self, rscpath):

        return self.confbypath[rscpath]

    def _cnames(self, resource):

        return list(resource)

    def _params(self, resource, cname):

        return resource[cname].values()

    def _setconf(self, conf, resource, rscpath):

        self.confbypath[rscpath] = conf

    def rscpaths(self, path):

        return [path]


class ConfDriverTest(UTCase):
    """Test the conf driver class."""

    __driverclass__ = TestConfDriver

    def setUp(self):

        self.count = 5
        self.driver = self.__driverclass__()
        self.paths = ['test{0}'.format(count) for count in range(self.count)]
        self.conf = Configuration(
            Category(
                'A',
                melts=[
                    Parameter('a', value=0, ptype=int),  # a is 0
                    Parameter('b', value=True, ptype=bool)
                ]
            ),  # b is overriden
            Category(
                'B',
                melts=[
                    Parameter('b', value=1, ptype=int),  # b is 1
                ]
            )
        )

    def test_rscpaths(self):
        """Test the method rscpaths."""

        for path in self.paths:

            rscpaths = self.driver.rscpaths(path=path)

            for rscpath in rscpaths:
                self.assertTrue(rscpath.endswith(path))

    def test_scenario(self):
        """Test a scenario of getting/putting configuration."""

        for path in self.paths:

            conf = self.driver.getconf(path=path)

            self.assertIsNone(conf)

            conf = self.driver.getconf(path=path, conf=self.conf)

            self.assertEqual(conf, self.conf)

            rscpaths = self.driver.rscpaths(path=path)

            for rscpath in rscpaths:

                self.assertTrue(rscpath.endswith(path))

                conf = self.conf.copy()
                conf += Category('test', melts=[Parameter('test')])

                self.driver.setconf(rscpath=rscpath, conf=conf)

            conf = self.driver.getconf(path=path)

            self.assertNotEqual(conf, self.conf)

            self.assertIn('test', conf)
            self.assertIn('test', conf['test'])


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""model.param UTs."""

from __future__ import absolute_import

from unittest import main

from six import string_types

from b3j0f.utils.ut import UTCase

from ..param import Parameter, PType, BOOL, ARRAY
from parser import ParserError


class PTypeTest(UTCase):
    """Test PType."""

    def setUp(self):

        self.ptype = PType(ptype=Parameter)

    def test_isinstance(self):
        """Test if isinstance."""

        self.assertIsInstance(Parameter(''), self.ptype)

    def test_PTypeisinstance(self):
        """Test isinstance of PTYpe."""

        self.assertIsInstance(self.ptype, self.ptype)

    def test_isnotinstance(self):
        """Test is not instance."""

        self.assertNotIsInstance(None, self.ptype)

    def test_issubclass(self):
        """Test is subclass."""

        self.assertTrue(issubclass(Parameter, self.ptype))

    def test_PTypeissubclass(self):
        """Test is subclass of PType."""

        self.assertTrue(issubclass(PType, self.ptype))

    def test_notissubclass(self):
        """Test is not subclass."""

        self.assertFalse(issubclass(type(None), self.ptype))

    def test_instanciate(self):
        """Test to instanciate a parameter."""

        value = 'test'

        param = self.ptype(value)

        self.assertIsInstance(param, Parameter)
        self.assertEqual(value, param.name)

    def test_instanciate_kwargs(self):
        """Test to instanciate a parameter with kwargs."""

        value = {'name': 'test'}

        param = self.ptype(value)

        self.assertIsInstance(param, Parameter)
        self.assertEqual(value['name'], param.name)

    def test_instanciate_args(self):
        """Test to instanciate a parameter with args."""

        value = ['test']

        param = self.ptype(value)

        self.assertIsInstance(param, Parameter)
        self.assertEqual(value[0], param.name)

    def test_instanciate_object(self):
        """Test to instanciate a parameter with an object such as name."""

        svalue = 1

        value = PType(str)(svalue)

        self.assertIsInstance(value, str)
        self.assertEqual(value, str(1))

    def test_instanciateerror(self):
        """Test to instanciate with an error."""

        self.assertRaises(ParserError, self.ptype, {'': None})


class BoolTest(UTCase):
    """Test the bool class."""

    def test(self):
        """Test to auto convert a value from BOOL."""

        param = Parameter('', svalue='false', ptype=BOOL)

        self.assertFalse(param.value)

        param.svalue = 'true'

        self.assertTrue(param.value)


class ArrayTest(UTCase):
    """Test the array class"""

    def test(self):
        """Test to auto convert a value from ARRAY."""

        param = Parameter('', svalue='', ptype=ARRAY)

        self.assertEqual(param.value, [])

        param = Parameter('', svalue='a', ptype=ARRAY)

        self.assertEqual(param.value, ['a'])

        param = Parameter('', svalue='a,b', ptype=ARRAY)

        self.assertEqual(param.value, ['a', 'b'])


class ParameterTest(UTCase):
    """Test a parameter."""

    def setUp(self):

        self.param = Parameter('test')

    def test___eq__strstr(self):
        """Test parameter comparison between str."""

        param = self.param.copy()

        self.assertIsInstance(param.name, string_types)
        self.assertEqual(param, self.param)

        param.name = param.name.upper()

        self.assertIsInstance(param.name, string_types)
        self.assertNotEqual(param, self.param)

    def test___eq__restr(self):
        """Test parameter comparison between regex and string."""

        param = Parameter('tes.')

        self.assertNotIsInstance(param.name, string_types)

        self.assertEqual(self.param, param)
        self.assertEqual(param, self.param)

        param.name = self.param.name + '.'

        self.assertNotIsInstance(param.name, string_types)
        self.assertNotEqual(self.param, param)
        self.assertNotEqual(param, self.param)

    def test___eq__rere(self):
        """Test parameter comparison between two regex."""

        self.param.name = '.*'
        param = self.param.copy()

        self.assertNotIsInstance(param.name, string_types)
        self.assertEqual(self.param, param)

        param.name = '.'
        self.assertNotIsInstance(param.name, string_types)
        self.assertNotEqual(self.param, param)

    def test_hash(self):
        """Test the method hash."""

        param = Parameter('test', )

        self.assertEqual(hash(param), hash(self.param))

        paramset = set((self.param, self.param.copy(), param))

        self.assertEqual(len(paramset), 1)

    def test_repr(self):
        """Test the method repr."""

        paramrepr = repr(self.param)
        cparamrepr = repr(self.param.copy())

        self.assertEqual(cparamrepr, paramrepr)

    def test_conf_name(self):
        """Test the method conf_name."""

        conf_name = self.param.conf_name

        self.assertEqual(
            conf_name, '{0}{1}'.format(self.param.name, Parameter.CONF_SUFFIX)
        )

    def test_svalue(self):
        """Test the property svalue."""

        self.assertIsNone(self.param.svalue)

        # check if parameter clean its protected attributes
        self.param._value = 1
        self.param._error = 1
        self.param.svalue = ''

        self.param.value = None

        self.assertIsNone(self.param._error)
        self.assertIsNone(self.param._value)
        self.assertEqual(self.param.svalue, '')

    def test_svalue_from_value(self):
        """Test to set svalue from value."""

        self.assertIsNone(self.param.svalue)

        # check if parameter clean it protected attributes
        self.param.value = '1'

        self.assertEqual(self.param.svalue, self.param.value)

    def test_svalue_nonify(self):
        """Test to nonify svalue."""

        self.param.value = 1
        self.param._error = 1
        self.param.svalue = None

        self.assertIsNotNone(self.param._value)
        self.assertIsNotNone(self.param._error)

        # nonify only if svalue exist
        self.param.svalue = '=1'
        self.assertIsNone(self.param._value)
        self.assertIsNone(self.param._error)

    def test_value(self):
        """Test the property value."""

        self.assertIsNone(self.param.value)

        self.param.value = 1

        self.assertEqual(self.param.value, 1)

    def test_value_from_svalue(self):
        """Test to set value from svalue."""

        self.param.svalue = '=1'

        value = self.param.value

        self.assertEqual(value, 1)

    def test_value_error_ptype(self):
        """Test to set value with wrong type."""

        self.param.ptype = int

        self.assertRaises(
            TypeError, lambda x: setattr(self.param, 'value', x), '3'
        )

        self.assertIsInstance(self.param.error, TypeError)

        self.param.svalue = ''

        self.assertRaises(Parameter.Error, lambda: getattr(self.param, 'value'))

    def test_copy(self):
        """Test the method copy."""

        param = Parameter(
            name='test',
            local=not Parameter.DEFAULT_LOCAL,
            ptype=int,
            conf=1,
            value=1,
            svalue='=1'
        )

        cparam = param.copy()

        self.assertEqual(param.name, cparam.name)
        self.assertEqual(param.local, cparam.local)
        self.assertEqual(param.ptype, cparam.ptype)
        self.assertEqual(param.conf, cparam.conf)
        self.assertEqual(param.value, cparam.value)
        self.assertEqual(param.svalue, cparam.svalue)

    def test_copy_cleaned(self):
        """Test the method copy with cleaned as True."""

        param = Parameter(
            name='test',
            local=not Parameter.DEFAULT_LOCAL,
            ptype=int,
            conf=1,
            value=1,
            svalue='=1'
        )

        cparam = param.copy(value=None, svalue=None)

        self.assertEqual(param.name, cparam.name)
        self.assertEqual(param.local, cparam.local)
        self.assertEqual(param.ptype, cparam.ptype)
        self.assertEqual(param.conf, cparam.conf)
        self.assertIsNone(cparam.value)
        self.assertIsNone(cparam.svalue)

    def test_simple_conversion(self):
        """Test simple conversion of parameter."""

        param = Parameter('', svalue='=1.0', ptype=int)

        self.assertIsInstance(param.value, int)

        param = Parameter('', svalue='=1.0')

        self.assertIsInstance(param.value, float)

if __name__ == '__main__':
    main()




__all__ = [
	'Configuration', 'configuration',
	'Category', 'category',
	'Parameter'
]

from .conf import Configuration, configuration
from .cat import Category, category
from .param import Parameter, BOOL, ARRAY, PType



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Javascript expression resolver."""

from __future__ import print_function

__all__ = ['resolvejs']

from ..registry import register

from ..core import DEFAULT_TOSTR, DEFAULT_SCOPE

try:
    from PyV8 import JSContext

except ImportError:

    from sys import stderr
    print(
        'Impossible to load the javascript resolver. Install the PyV8 before.',
        file=stderr
    )
    def resolvejs(**_):
        """Default resolvejs if PyV8 is not installed."""
        pass

else:
    CTXT = JSContext()

    @register('js')
    @register('javascript')
    @register('pyv8')
    def resolvejs(
            expr, tostr=DEFAULT_TOSTR, scope=DEFAULT_SCOPE, *args, **kwargs
    ):
        """Javascript resolver."""

        _ctxt = CTXT if scope is None else JSContext(scope)

        if tostr:
            expr = '({0}).string'.format(expr)

        result = _ctxt.eval(expr)

        return result




# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Specification of the class Configurable."""

__all__ = ['Configurable']

from six import string_types
from six.moves import reload_module

from inspect import getargspec, isclass

from traceback import format_exc

from b3j0f.utils.path import lookup
from b3j0f.utils.version import getcallargs
from b3j0f.annotation import PrivateInterceptor, Annotation

from ..model.conf import Configuration, configuration
from ..model.cat import Category, category
from ..model.param import Parameter
from ..driver.file.json import JSONFileConfDriver
from ..driver.file.ini import INIFileConfDriver
from ..driver.file.xml import XMLFileConfDriver
from ..parser.resolver.core import (
    DEFAULT_SAFE, DEFAULT_SCOPE, DEFAULT_BESTEFFORT
)

from random import random


class Configurable(PrivateInterceptor):
    """Handle object configuration from configuration resources such as files.

    Based on a configuration model and configuration drivers, it is able to:

    - be used such as a decorator (see the b3j0f.annotation.Annotation). resources to configure are accessible from the
        ``targets`` property.
    - inject configuration parameters at call time (for function/method calls).
    - preload required modules.
    - be used with several objects.

    Usage :

    .. code-block:: python

        from b3j0f.conf import category
        from b3j0f.conf import Configurable
        from b3j0f.conf getconfigurables, applyconfiguration

        # class configuration
        @Configurable(conf=category('land', Parameter('country', value='fr')))
        class World(object):
            pass

        world = World()
        assert world.country == 'fr'

        # reuse the configuration on several objects
        land = getconfigurables(world)[0]

        @land
        class World2(object):
            pass

        assert World2().country == 'fr'

        land.conf['land']['country'].value = 'en'

        # update all configuration on an object
        applyconfiguration([world])

        assert world.country == 'en'

        # inject configuration in the constructor and on object
        class World3(object):
            def __init__(self, country=None):
                self.safecountry = country

        world3 = World3()
        land(world3)

        assert world3.country == 'en'
        assert world3.safecountry = 'en'

        # inject configuration in function parameters
        @land
        def getcountry(country=None):
            print(country)
            return country

        assert getcountry() == 'en'
    """

    CATEGORY = 'CONFIGURABLE'  #: configuration category name.

    CONF = 'conf'  #: self configuration attribute name.
    PATHS = 'paths'  #: paths attribute name.
    DRIVERS = 'drivers'  #: drivers attribute name.
    FOREIGNS = 'foreigns'  #: not specified params setting attribute name.
    AUTOCONF = 'autoconf'  #: auto conf attribute name.
    SAFE = 'safe'  #: safe attribute name.
    SCOPE = 'scope'  #: scope attribute name.
    BESTEFFORT = 'besteffort'  #: best effort attribute name.
    MODULES = 'modules'  #: modules attribute name.
    CALLPARAMS = 'callparams'  #: call params attribute name.
    KEEPSTATE = 'keepstate'  #: reconfiguration keepstate level.

    EXEC_CTX = '__CONFIGURABLE_EXEC_CTX' #: configurable execution context.

    DEFAULT_FOREIGNS = True  #: default value for setting not specified params.
    # default drivers which are json and ini.
    DEFAULT_DRIVERS = (
        JSONFileConfDriver(), INIFileConfDriver(), XMLFileConfDriver()
    )
    DEFAULT_AUTOCONF = True  #: default value for auto configuration.
    DEFAULT_CALLPARAMS = True  #: default value for callparams.
    DEFAULT_MODULES = None  #: default value for modules.
    DEFAULT_SAFE = DEFAULT_SAFE  #: default value for safe.
    DEFAULT_SCOPE = DEFAULT_SCOPE  #: default value for scope.
    DEFAULT_BESTEFFORT = DEFAULT_BESTEFFORT  #: default value for besteffort.
    DEFAULT_KEEPSTATE = True  #: default keepstate value.

    SUB_CONF_PREFIX = ':'  #: sub conf prefix.

    def __init__(
            self,
            conf=None, paths=None, drivers=DEFAULT_DRIVERS,
            foreigns=DEFAULT_FOREIGNS, autoconf=DEFAULT_AUTOCONF, targets=None,
            keepstate=DEFAULT_KEEPSTATE,
            safe=DEFAULT_SAFE, scope=DEFAULT_SCOPE,
            besteffort=DEFAULT_BESTEFFORT, modules=DEFAULT_MODULES,
            callparams=DEFAULT_CALLPARAMS, logger=None,
            *args, **kwargs
    ):
        """
        :param conf: conf to use at instance level.
        :type conf: Configuration, Category or Parameter.
        :param Iterable paths: paths to parse.
        :param targets: object(s) to reconfigure. Such object may
            implement the methods configure applyconfiguration and configure.
        :param bool keepstate: if True (default), do not instanciate sub objects
            if they already exist.
        :type targets: list or instance.
        :param bool foreigns: if True (default), set parameters not specified by
            this conf but given by conf resources.
        :param ConfDriver(s) drivers: list of drivers to use. Default
            Configurable.DEFAULT_DRIVERS.
        :param list targets: objects to configure.
        :param bool autoconf: if autoconf, configurate this `targets`
            objects as soon as possible. (at targets instanciation or after
            updating this paths/conf/targets).
        :param bool safe: if True (default), expression parser are used in a
            safe context to resolve python object. For example, if safe,
            builtins function such as `open` are not resolvable.
        :param dict scope: expression resolver scope.

        :param bool besteffort: expression resolver best effort flag.
        :param list modules: required modules.
        :param bool callparams: if True (default), use parameters in the
            configured callable function.
        :param bool autoresolve: if True (default), call auto resolve on
            retrieved configurations.
        :param Logger logger: this logger."""

        super(Configurable, self).__init__(*args, **kwargs)

        # init protected attributes
        self._paths = None
        self._conf = None
        self._targets = set()
        self._modules = []

        # init public attributes

        # dirty hack: falsify autoconf in order to avoid auto applyconfiguration
        self.autoconf = False

        self.drivers = drivers
        self.foreigns = foreigns
        self.targets = set() if targets is None else targets
        self.keepstate = keepstate
        self.conf = conf
        self.paths = paths
        self.safe = safe
        self.scope = scope
        self.besteffort = besteffort
        self.modules = modules
        self.callparams = callparams
        self.logger = logger

        # generate an execution context name
        self.exec_ctx = '{0}{1}'.format(Configurable.EXEC_CTX, random())

        self.autoconf = autoconf  # end of dirty hack

        if self.autoconf:
            self.applyconfiguration()

    def _interception(self, joinpoint):

        target = joinpoint.target

        args, kwargs = joinpoint.args, joinpoint.kwargs

        if self.callparams:

            conf = self.getconf()

            args, kwargs = self.getcallparams(
                conf=conf, target=target, args=list(args), kwargs=kwargs,
                exec_ctx=joinpoint.exec_ctx.setdefault(self.exec_ctx, set())
            )

            joinpoint.args = list(args)
            joinpoint.kwargs = kwargs

        result = joinpoint.proceed()

        if self.autoconf:

            targets = [result]

            cls = joinpoint.ctx if isclass(joinpoint.ctx) else target

            if result is None:

                if isclass(cls):
                    # if target is the method __init__
                    if target.__name__ == '__init__':
                        if 'self' in kwargs:
                            targets = [kwargs['self']]

                        else:
                            targets = [args[0]]

            if not isclass(cls) or isinstance(targets[0], cls):
                self.applyconfiguration(targets=targets, callconf=False)

        return result

    def getcallparams(
            self, target, conf=None, args=None, kwargs=None, exec_ctx=None
    ):
        """Get target call parameters.

        :param list args: target call arguments.
        :param dict kwargs: target call keywords.
        :return: args, kwargs
        :rtype: tuple"""

        if args is None:
            args = []

        if kwargs is None:
            kwargs = {}

        if conf is None:
            conf = self.conf

        params = conf.params

        try:
            argspec = getargspec(target)

        except TypeError as tex:
            argspec = None
            callargs = {}

        else:
            try:
                callargs = getcallargs(target, *args, **kwargs)

            except TypeError as tex:
                if tex.args[0].endswith('\'{0}\''.format(argspec.args[0])):
                    args = [None]

                else:
                    raise

                callargs = getcallargs(target, *args, **kwargs)

                args = []

        pnames = set(params)

        for pname in pnames:

            if argspec and pname in argspec.args and (
                    (exec_ctx is not None and pname in exec_ctx)
                    or
                    callargs.get(pname) is None
            ):

                kwargs[pname] = params[pname].value

        if exec_ctx is not None:
            exec_ctx |= pnames

        return args, kwargs

    @property
    def modules(self):
        """Get this required modules."""

        return self._modules

    @modules.setter
    def modules(self, value):
        """Change required modules.

        Reload modules given in the value.

        :param list value: new modules to use."""

        self._modules = value

        if value:
            for module in value:
                reload_module(lookup(module))

    @property
    def conf(self):
        """Get conf with parsers and self property values.

        :rtype: Configuration.
        """

        return self._conf

    @conf.setter
    def conf(self, value):
        """Change of configuration.

        :param value: new configuration to use.
        :type value: Category or Configuration
        """

        self._conf = self._toconf(value)

        if self.autoconf:
            self.applyconfiguration()

    def _toconf(self, conf):
        """Convert input parameter to a Configuration.

        :param conf: configuration to convert to a Configuration object.
        :type conf: Configuration, Category or Parameter.

        :rtype: Configuration"""

        result = conf

        if result is None:
            result = Configuration()

        elif isinstance(result, Category):
            result = configuration(result)

        elif isinstance(result, Parameter):
            result = configuration(category('', result))

        return result

    @property
    def paths(self):
        """Get all type conf files and user files.

        :return: self conf files
        :rtype: tuple
        """

        return tuple(self._paths)

    @paths.setter
    def paths(self, value):
        """Change of paths in adding it in watching list."""

        if value is None:
            value = ()

        elif isinstance(value, string_types):
            value = (value, )

        self._paths = tuple(value)

        if self.autoconf:
            self.applyconfiguration()

    def applyconfiguration(
            self, conf=None, paths=None, drivers=None, logger=None,
            targets=None, scope=None, safe=None, besteffort=None,
            callconf=False, keepstate=None
    ):
        """Apply conf on a destination in those phases:

        1. identify the right driver to use with paths to parse.
        2. for all paths, get conf which matches input conf.
        3. apply parsing rules on path parameters.
        4. fill input conf with resolved parameters.
        5. apply filled conf on targets.

        :param conf: conf from where get conf.
        :type conf: Configuration, Category or Parameter
        :param paths: conf files to parse. If paths is a str, it is
            automatically putted into a list.
        :type paths: list of str
        :param Iterable targets: objects to configure. this targets by default.
        :param dict scope: local variables to use for expression resolution.
        :param bool callconf: if True (False by default), the configuration is
            used in the callable target parameters while calling it.
        :param bool keepstate: if True (default), do not instanciate sub objects
            if they already exist.
        :return: configured targets.
        :rtype: list
        """

        result = []

        if scope is None:
            scope = self.scope

        if safe is None:
            safe = self.safe

        if besteffort is None:
            besteffort = self.besteffort

        if targets is None:
            targets = self.targets

        if logger is None:
            logger = self.logger

        if keepstate is None:
            keepstate = self.keepstate

        conf = self._toconf(conf)

        # get conf from drivers and paths
        conf = self.getconf(
            conf=conf, paths=paths, logger=logger, drivers=drivers
        )
        if conf is not None:
            # resolve all values
            conf.resolve(
                configurable=self,
                scope=scope, safe=safe, besteffort=besteffort
            )
            # configure resolved configuration
            configured = self.configure(
                conf=conf, targets=targets, callconf=callconf,
                keepstate=keepstate
            )

            result += configured

        return result

    def getconf(self, conf=None, paths=None, drivers=None, logger=None):
        """Get a configuration from paths.

        :param conf: conf to update. Default this conf.
        :type conf: Configuration, Category or Parameter
        :param str(s) paths: list of conf files. Default this paths.
        :param Logger logger: logger to use for logging info/error messages.
        :param list drivers: ConfDriver to use. Default this drivers.
        :return: not resolved configuration.
        :rtype: Configuration
        """

        result = None

        conf = self._toconf(conf)

        # start to initialize input params
        if conf is None:
            conf = self.conf.copy()

        else:
            selfconf = self.conf.copy()
            selfconf.update(conf)
            conf = selfconf

        if paths is None:
            paths = self.paths

        if isinstance(paths, string_types):
            paths = [paths]

        if drivers is None:
            drivers = self.drivers

        if logger is None:
            logger = self.logger

        # iterate on all paths
        for path in paths:

            for driver in drivers:  # find the best driver

                rscconf = driver.getconf(
                    path=path, conf=conf, logger=logger
                )

                if rscconf is None:
                    continue

                if result is None:
                    result = rscconf

                else:
                    result.update(rscconf)

            if result is None:
                # if no conf found, display a warning log message
                if logger is not None:
                    logger.warning(
                        'No driver found among {0} for processing {1}'.format(
                            drivers, path
                        )
                    )

        return result

    def configure(
            self, conf=None, targets=None, logger=None, callconf=False,
            keepstate=None
    ):
        """Apply input conf on targets objects.

        Specialization of this method is done in the _configure method.

        :param conf: configuration model to configure. Default is this conf.
        :type conf: Configuration, Category or Parameter
        :param Iterable targets: objects to configure. self targets by default.
        :param Logger logger: specific logger to use.
        :param bool callconf: if True (False by default), the configuration is
            used in the callable target parameters while calling it.
        :param bool keepstate: if True (default), do not instanciate sub objects
            if they already exist.
        :return: configured targets.
        :rtype: list
        :raises: Parameter.Error for any raised exception.
        """

        result = []

        conf = self._toconf(conf)

        if conf is None:
            conf = self.conf

        if targets is None:
            targets = self.targets

        if logger is None:
            logger = self.logger

        if keepstate is None:
            keepstate = self.keepstate

        for target in targets:

            try:
                configured = self._configure(
                    conf=conf, logger=logger, target=target, callconf=callconf,
                    keepstate=keepstate
                )

            except:
                if logger is not None:
                    logger.error(
                        'Error {0} raised while configuring {1}/{2}'.format(
                            format_exc(), self, targets
                        )
                    )

            else:
                result.append(configured)

        return result

    def _configure(
            self, target, conf=None, logger=None, callconf=None, keepstate=None
    ):
        """Configure this class with input conf only if auto_conf or
        configure is true.

        This method should be overriden for specific conf

        :param target: object to configure. self targets by default.
        :param Configuration conf: configuration model to configure. Default is
            this conf.
        :param Logger logger: logger to use.
        :param bool callconf: if True, use conf in target __call__ parameters.
        :param bool keepstate: if True recreate sub objects if they already
            exist.
        """

        result = target

        if conf is None:
            conf = self.conf

        if logger is None:
            logger = self.logger

        if callconf is None:
            callconf = self.callparams

        if keepstate is None:
            keepstate = self.keepstate

        subcats = {}  # store sub configurable categories
        params = []  # self parameters

        sub_conf_prefix = Configurable.SUB_CONF_PREFIX

        for cat in conf.values():  # separate sub params and params
            cname = cat.name

            if cname.startswith(sub_conf_prefix):

                subcnames = cname.split(sub_conf_prefix)

                pname = subcnames[1]
                fcname = cname[1 + len(pname):]

                if not fcname:
                    fcname = str(random())

                fcat = cat.copy(name=fcname)

                if pname in subcats:
                    subcats[pname].append(fcat)

                else:
                    subcats[pname] = [fcat]

            else:
                cparams = cat.params
                params += cparams.values()

        if callconf and callable(target):
            conf = configuration(category('', *params))

            args, kwargs = self.getcallparams(conf=conf, target=target)

            result = target = target(*args, **kwargs)

        for param in params:

            value, pname = param.value, param.name

            if pname in subcats:  # if sub param

                subcallconf = True

                if keepstate and hasattr(target, pname):
                    subcallconf = False
                    value = getattr(target, pname)

                cats = subcats[pname]

                subconf = configuration(*cats)

                value = applyconfiguration(
                    targets=[value], conf=subconf, callconf=subcallconf,
                    keepstate=keepstate
                )[0]

            if param.error:
                continue

            elif self.foreigns or param.local:
                try:
                    setattr(target, pname, value)

                except Exception:
                    if logger is not None:
                        logger.error(
                           'Error while setting {0}({1}) on {2}: {3}'.format(
                            pname, value, target, format_exc()
                        )
                    )

        return result

    def _bind_target(self, target, ctx, *args, **kwargs):

        if callable(target):

            try:
                result = super(Configurable, self)._bind_target(
                    target=target, ctx=ctx, *args, **kwargs
                )

            except AttributeError:
                self.remove_from(target=target, ctx=ctx)
                def __init__(*args, **kwargs):
                    return object.__init__(*args, **kwargs)
                target.__init__ = __init__
                result = super(Configurable, self)._bind_target(
                    target=target, ctx=ctx, *args, **kwargs
                )

        else:
            result = Annotation._bind_target(
                self, target=target, ctx=ctx, *args, **kwargs
            )

        return result


def applyconfiguration(targets, conf=None, *args, **kwargs):
    """Apply configuration on input targets.

    If targets are not annotated by a Configurable, a new one is instanciated.

    :param Iterable targets: targets to configurate.
    :param tuple args: applyconfiguration var args.
    :param dict kwargs: applyconfiguration keywords.
    :return: configured targets.
    :rtype: list
    """

    result = []

    for target in targets:

        configurables = Configurable.get_annotations(target)

        if not configurables:
            configurables = [Configurable()]

        for configurable in configurables:

            configuredtargets = configurable.applyconfiguration(
                targets=[target], conf=conf, *args, **kwargs
            )

            result += configuredtargets

    return result

#: Default Configurable conf.
_CONF = configuration(
    category(
        Configurable.CATEGORY,
        Parameter(
            name=Configurable.DRIVERS, ptype=tuple,
            value=Configurable.DEFAULT_DRIVERS
        ),
        Parameter(
            name=Configurable.PATHS, ptype=tuple,
            value=('b3j0fconf-configurable.conf', )
        ),
        Parameter(
            name=Configurable.SCOPE, ptype=dict,
            value=Configurable.DEFAULT_SCOPE
        ),
        Parameter(
            name=Configurable.SAFE, ptype=bool,
            value=Configurable.DEFAULT_SAFE
        ),
        Parameter(
            name=Configurable.BESTEFFORT, ptype=bool,
            value=Configurable.DEFAULT_BESTEFFORT
        ),
        Parameter(
            name=Configurable.CALLPARAMS, ptype=bool,
            value=Configurable.DEFAULT_CALLPARAMS
        ),
        Parameter(
            name=Configurable.MODULES, ptype=tuple,
            value=Configurable.DEFAULT_MODULES
        ),
        Parameter(
            name=Configurable.AUTOCONF, ptype=bool,
            value=Configurable.DEFAULT_AUTOCONF
        ),
        Parameter(
            name=Configurable.FOREIGNS, ptype=bool,
            value=Configurable.DEFAULT_FOREIGNS
        ),
        Parameter(
            name=Configurable.KEEPSTATE, ptype=bool,
            value=Configurable.DEFAULT_KEEPSTATE
        )
    )
)

# apply the configurable on itself
Configurable(conf=_CONF)(Configurable)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Programming language expression resolver module.

This module defines expression resolvers for dedicated programming languages.

An expression resolver is a function which takes in parameters:

- expr: string expression to resolve.
- safe: boolean flag for a safe execution context.
- tostr: boolean flag about format the expression into a string representation.
- scope: dict of variables for execution context (contain variables).

A resolver is registered with the function ``register``.
And is loaded in three ways:

- in setting the environment variable 'B3J0F_EXPRRES_PATH' (for example,
    `b3j0f.conf.parser.lang.js,custom.c` could load both modules `js` and `c`
    containing programming language parsers).
- in using the function `loadresolvers`.
- in simply importing a dedicated module with the import keyword.
"""

__all__ = ['ExprResolver']

from six import add_metaclass

from .registry import register

from .core import (
    DEFAULT_BESTEFFORT, DEFAULT_SAFE, DEFAULT_TOSTR, DEFAULT_SCOPE
)


class _MetaExprResolver(type):
    """Expression Resolver meta class.

    Register automatically ExprResolver classes."""

    def __new__(cls, *args, **kwargs):

        result = super(_MetaExprResolver, cls).__new__(cls, *args, **kwargs)

        if result.__register__:
            register(exprresolver=result)

        return result


@add_metaclass(_MetaExprResolver)
class ExprResolver(object):
    """Expression resolver class.

    All sub classes are automatically registered."""

    __register__ = False  #: class registration flag.

    def __call__(
            self, expr,
            safe=DEFAULT_SAFE, tostr=DEFAULT_TOSTR, scope=DEFAULT_SCOPE,
            besteffort=DEFAULT_BESTEFFORT
    ):
        """Resolve input expression.

        :param str expr: configuration expression to resolve in this language.
        :param bool safe: safe run execution context (True by default).
        :param bool tostr: format the result.
        :param dict scope: execution scope (contains references to expression
            objects).
        :param bool besteffort: try to resolve unknown variable name from
            runtime context.
        """

        raise NotImplementedError()
